"""beaker-jobwatch - monitor and reschedule beaker jobs."""

import datetime
import functools
import optparse
import os
import pathlib
import re
import select
import signal
import subprocess
import sys
import tempfile
import time
import typing
import xmlrpc.client

import lxml.etree
import requests

__version__ = "8.0"

STATUS_DOT_COUNT = 30

OPTIONS: optparse.Values

XMLRPC_CLIENT = xmlrpc.client.ServerProxy("https://beaker.engineering.redhat.com")
SESSION = requests.Session()


def read_options(
    param_options,
    ignore_list,
):
    """Read jobs to run."""
    herd = []
    rs_list = []
    duration = {}

    # populate herd with all recipesets from all given jobs (from both test-run and cmdline)
    for job in set(re.split(r"\W+", ",".join(param_options.jobs))):
        # if input was in TJ#XYZ format, if was split as [TJ,XYZ] - skip the TJ part
        # ditto for J:XYZ and empty jobs
        if job in {"TJ", "J", ""}:
            continue
        print(f"processing job TJ#{job}")
        recipe_sets = RecipeSetCache.add_job(job)
        for rs_id in recipe_sets:
            # check whether was the recipeSet already processed
            if rs_id in rs_list:
                if OPTIONS.debug:
                    print(f"already processed recipeSet {rs_id}")
            else:
                if OPTIONS.debug:
                    print(f"watching recipeSet {rs_id}")
                rs = RecipeSet(
                    rs_id,
                    job,
                    param_options.max_reschedules,
                    param_options.max_reschedules,
                    param_options.skip_destructive,
                    param_options.skip_machine,
                    ignore_list,
                )
                if rs.is_nacked() and not param_options.load_nacked:
                    if OPTIONS.debug:
                        print("recipeSet already NACKed, skipping")
                else:
                    # first recipe for this arch, note the time
                    if rs.description not in duration:
                        duration[rs.description] = {
                            "start": datetime.datetime.now(datetime.UTC),
                            "end": None,
                        }
                    herd.append(rs)
                    # add RS ID to list of already parsed RS
                    rs_list.append(rs_id)
                    if OPTIONS.debug:
                        print(f"newly processed recipeSet {rs_id}")

    if len(herd) == 0:
        print("no recipeSets found (all nacked?)")
        sys.exit(1)

    print(f"\nThere was {len(rs_list)} new recipeSet(s) added")

    return len(rs_list), herd, duration


class AlarmError(Exception):
    """support class/function for timeouting bkr calls."""

    @staticmethod
    def alarm_handler(_signum, _frame):
        """Raise an AlarmError."""
        raise AlarmError()


class Bkr:
    """Wrapper for bkr cmdline client calls."""

    username = None
    password = None

    class MaxRetriesReachedError(Exception):
        """Raised when command failed too many times."""

    @classmethod
    def run(cls, args):
        """Run a bkr command and return the process handle."""
        cmd = ["bkr"]
        cmd.append(args[0])
        if cls.username and cls.password:
            cmd.extend([f"--username={cls.username}", f"--password={cls.password}"])
        cmd.extend(args[1:])
        return subprocess.Popen(args=cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    @classmethod
    def timeout(cls, process, timeout=600):
        """Wait for process to finish or timeout after timeout seconds."""
        out = None
        err = None
        signal.signal(signal.SIGALRM, AlarmError.alarm_handler)
        signal.alarm(timeout)
        try:
            (out, err) = process.communicate()
            signal.alarm(0)  # reset alarm
        except AlarmError:
            if OPTIONS.debug:
                print(f"timeout of {timeout} reached, terminating process {process.pid}")
            try:
                process.terminate()
                time.sleep(5)
                if process.poll() is None:
                    process.kill()
            except Exception:
                if OPTIONS.debug:
                    print("terminating process failed")
            time.sleep(10)
        return (out, err)

    @classmethod
    def run_repeatedly(cls, command, max_retries=5, pause=60):
        """Run command until it exits with 0 or we run out of retries."""
        retries = 0
        while retries < max_retries:
            p = cls.run(command)
            (output, error) = Bkr.timeout(p)
            if p.wait() == 0:
                return (output, error)
            print(f"bkr command {command} failed (try #{retries}):")
            print(error)
            time.sleep(pause)
            pause += 60
            retries += 1
        raise Bkr.MaxRetriesReachedError()


class LogCache:
    """Stores log content from beaker recipes."""

    logs: typing.ClassVar = {}
    log_list: typing.ClassVar = {}

    @classmethod
    def get_log(cls, log_url):
        """Return content of one log identified by a log URL (from cache if available)."""
        # check cache, update if empty
        if log_url not in cls.logs:
            if OPTIONS.debug:
                print(f"downloading log {log_url}")
            response = SESSION.get(log_url)
            response.raise_for_status()
            cls.logs[log_url] = response.content.decode("utf-8")
        return cls.logs[log_url]

    @classmethod
    def get_recipe_logs(cls, recipe_id, log_name="install.log"):
        """Return content of all logs for a particular recipe that matches log_name."""
        # check cache, update if empty
        if recipe_id not in cls.log_list:
            if OPTIONS.debug:
                print(f"getting log list for recipe R:{recipe_id}")
            p = Bkr.run(["job-logs", f"R:{recipe_id}"])
            (output, err) = Bkr.timeout(p)
            output = output.decode()
            err = err.decode()
            if p.wait() != 0:
                print(f"retrieving recipe log list failed for R:{recipe_id}")
                return []
            cls.log_list[recipe_id] = list(str(output).split("\n"))
        log_list = [x for x in cls.log_list[recipe_id] if log_name in x]
        logs = []
        for log_url in log_list:
            try:
                logs.append(cls.get_log(log_url))
            except Exception as e:
                print(f"failed to retrieve log {log_url}: {e}")
        return logs


class RecipeSetCache:
    """Stores recipeset results in a parsed xml as returned by bkr job-results.

    Automatically updates cache for running RSs if older than defined interval.
    """

    cache: typing.ClassVar = {}
    update_interval = 60

    @classmethod
    def add(cls, rs_id):
        """Add a recipeSet to the cache."""
        xml = cls._get_results(rs_id)
        # keep whiteboard if we already have it
        try:
            job_whiteboard = cls.cache[rs_id]["job_whiteboard"]
        except KeyError:
            job_whiteboard = "Unknown"
        try:
            first_seen = cls.cache[rs_id]["first_seen"]
        except KeyError:
            first_seen = datetime.datetime.now(tz=datetime.UTC)
        cls.cache[rs_id] = {
            "xml": xml,
            "time": time.time(),
            "job_whiteboard": job_whiteboard,
            "first_seen": first_seen,
        }

    @classmethod
    def add_job(cls, job_id):
        """Add all recipeSets of a job to the cache.

        Returns list of recipeSet IDs found.
        """
        xml = cls._get_results(job_id, recipe_set=False)
        job_whiteboard = xml.find("whiteboard").text
        rs_ids = []
        for rs in xml.findall(".//recipeSet"):
            rs_id = rs.get("id")
            rs_ids.append(rs_id)
            try:
                first_seen = cls.cache[rs_id]["first_seen"]
            except KeyError:
                first_seen = datetime.datetime.now(tz=datetime.UTC)
            cls.cache[rs_id] = {
                "xml": rs,
                "time": time.time(),
                "job_whiteboard": job_whiteboard,
                "first_seen": first_seen,
            }
        return rs_ids

    @classmethod
    def get(
        cls,
        rs_id,
        *,
        force: bool = False,
        cache_only: bool = False,
    ) -> lxml.etree._Element | None:
        """Return XML for a recipeSet from the cache."""
        if rs_id not in cls.cache:
            return None
        if not force:
            # results are still fresh
            if (time.time() - cls.cache[rs_id]["time"]) < cls.update_interval:
                return cls.cache[rs_id]["xml"]
            if cache_only:
                return cls.cache[rs_id]["xml"]
        # RS is finished, no update necessary
        running = False
        for recipe in cls.cache[rs_id]["xml"].findall(".//recipe"):
            if recipe.get("status") not in ("Completed", "Cancelled", "Aborted"):
                running = True
        if running:
            cls.add(rs_id)
        return cls.cache[rs_id]["xml"]

    # an alias for the get method
    update = get

    @classmethod
    def _get_results(cls, job_or_rs_id, *, recipe_set: bool = True) -> lxml.etree._Element:
        """Get the XML results of a job/recipeSet from Beaker."""
        prefix = "RS"
        if not recipe_set:
            prefix = "J"
        retries = 0
        while True:
            if OPTIONS.debug:
                print(f"fetching xml results of {prefix}:{job_or_rs_id}")
            p = Bkr.run(["job-results", "--no-logs", f"{prefix}:{job_or_rs_id}"])
            (output, _err) = Bkr.timeout(p)
            if p.wait() == 0:
                break
            print(
                f"retrieving {prefix} results failed for {prefix}:{job_or_rs_id} (try #{retries})",
            )
            retries += 1
            time.sleep(120)
        parser = lxml.etree.XMLParser(remove_blank_text=True)
        return lxml.etree.XML(output, parser)

    @classmethod
    def first_seen(cls, rs_id):
        """Return timestamp of when recipeSet was added to the cache."""
        return cls.cache[rs_id]["first_seen"]

    @classmethod
    def whiteboard(cls, rs_id):
        """Return job whiteboard for this RS."""
        return cls.cache[rs_id]["job_whiteboard"]

    @classmethod
    def dump_to_disk(cls, rs_id):
        """Dump the XML of a recipeSet to disk."""
        (fd, filename) = tempfile.mkstemp()
        file = os.fdopen(fd, "wb")
        file.write(lxml.etree.tostring(cls.cache[rs_id]["xml"]))
        file.close()
        print(f"RS {rs_id} dumped to {filename}")


class RecipeSet:
    """Represents a recipeSet in Beaker."""

    def __init__(
        self,
        rs_id,
        job,
        reschedules,
        max_reschedules,
        skip_destructive,
        skip_machine,
        ignore_list=None,
    ):
        """Create a new RecipeSet object."""
        if ignore_list is None:
            ignore_list = []
        self.id = rs_id
        self.job = job
        self.reschedules_left = reschedules
        self.max_reschedules = max_reschedules
        self.status_description = None
        self.active = True
        self.skip_destructive = skip_destructive
        self.skip_machine = skip_machine
        self.description = ""
        self.ignore_list = ignore_list
        xml = self.get_xml()
        for recipe in xml.findall(".//recipe"):
            self.description = f"{self.description},{recipe.get('arch')}"
        self.description = self.description.lstrip(",")
        if OPTIONS.debug:
            print(
                f"new recipeSet created (id={rs_id}, job={job}, "
                f"reschedules_left={reschedules},ignore_list={ignore_list})",
            )
        self.last_ping = {}
        self.last_reboot = {}
        self.reboot_time = 60 * OPTIONS.reboot_time
        self.cur_reboots_without_ping = {}

    def __str__(self):
        """Return the recipeSet ID as the string representation."""
        return str(self.id)

    def get_xml(self, *, cache_only: bool = True) -> lxml.etree._Element:
        """Return the parsed XML result of the entire RS."""
        if OPTIONS.aggressive:
            cache_only = False
        return RecipeSetCache.get(self.id, cache_only=cache_only)

    def is_ok(self):
        """Return whether no error was found (so far) in this recipeSet.

        Equals to is_finished unless in aggressive mode in which case
        the full results XML is fetched and analyzed.
        """
        # queued for longer than allowed
        if OPTIONS.queue_timeout:
            task_info = self.get_task_info()
            if task_info and task_info["state"] == "Queued":
                tracked_for = (
                    datetime.datetime.now(tz=datetime.UTC) - RecipeSetCache.first_seen(self.id)
                ).total_seconds()
                if OPTIONS.debug:
                    print(
                        f"RS:{self.id} first seen {RecipeSetCache.first_seen(self.id)}, "
                        f"remaining life: {OPTIONS.queue_timeout - tracked_for}",
                    )
                if tracked_for > OPTIONS.queue_timeout:
                    self.status_description = "queue timeout"
                    return False

        # reboots
        if OPTIONS.reboot:
            for box in self.get_machines():
                if subprocess.run(["ping", "-c", "1", box], check=False).returncode == 0:
                    if OPTIONS.debug:
                        print(f"Successful ping to {box} ( J:{self.job} / RS:{self.id} )")
                    self.last_ping[box] = time.time()
                    self.cur_reboots_without_ping[box] = 0
                else:
                    if (box in self.last_reboot) and (
                        (time.time() - self.last_reboot[box]) < self.reboot_time
                    ):
                        if OPTIONS.debug:
                            print(f"Waiting for {box} to reboot ( J:{self.job} / RS:{self.id} )")
                        continue
                    if OPTIONS.debug:
                        print(f"Unsuccessful ping to {box} ( J:{self.job} / RS:{self.id} )")
                # consider rebooting
                if box not in self.last_ping:
                    # here we assume that the box didn't install yet
                    if OPTIONS.debug:
                        print(
                            f"Not a single ping to {box} yet ( J:{self.job} / RS:{self.id} ). "
                            f"Skipping it for now.",
                        )
                    continue
                if time.time() - self.last_ping[box] > self.reboot_time:
                    if self.cur_reboots_without_ping[box] >= OPTIONS.max_reboots_without_ping:
                        print(
                            f"System {box} fails to ping after "
                            f"{self.cur_reboots_without_ping[box]} reboots, rescheduling "
                            f"J:{self.job} / RS:{self.id}.",
                        )
                        # Cause a reschedule() and add the new job to herd (!)
                        self.status_description = f"failed to ping {box} after reboot"
                        return False
                    print(f"Rebooting {box} ( J:{self.job} / RS:{self.id} ) now ...")
                    subprocess.run(["bkr", "system-power", "--action", "reboot", box], check=False)
                    self.last_reboot[box] = time.time()
                    self.cur_reboots_without_ping[box] += 1
                    continue

        # to decrease load on Beaker, don't fetch full xml
        if not OPTIONS.aggressive and not self.is_finished():
            return True

        rs_xml = self.get_xml(cache_only=False)

        # destructive task aborting all tasks following it
        if ("destructive-task" not in self.ignore_list) and (
            destructive_tasks := self.get_destructive_tasks()
        ):
            self.status_description = f"destructive task(s) {
                functools.reduce(
                    lambda x, y: f'{x},{y}',
                    [task.get('name') for task in destructive_tasks],
                )
            }"
            return False

        # aborted/paniced recipes
        if "abort" not in self.ignore_list:
            for recipe in rs_xml.findall(".//recipe"):
                if recipe.get("status") == "Aborted":
                    # only check for panics in aborted recipes
                    # panics in completed ones are task-errors but not
                    # job breakers = ok, no need to reschedule
                    if recipe.get("result") == "Panic":
                        self.status_description = "kernel panic"
                    else:
                        self.status_description = "recipe aborted"
                        # check if no HW available for the recipe
                        if (results := recipe.xpath("//result")) and (
                            "does not match any systems" in results[0].text
                        ):
                            self.status_description = "no systems match required HW"
                            # irrecoverable error, don't reschedule
                            self.reschedules_left = 0
                        # check install logs for common errors
                        # check for broken scriptlets
                        logs = LogCache.get_recipe_logs(
                            recipe_id=recipe.get("id"),
                            log_name="install.log",
                        )
                        if len(logs) > 0:
                            for line in logs[0].split("\n"):
                                match = re.match(r"^error:.*\((.*)\) scriptlet failed.*", line)
                                if match:
                                    package_name = match.group(1)
                                    self.status_description = (
                                        f"bad scriptlet in {package_name} package"
                                    )
                                    # irrecoverable error, don't reschedule
                                    self.reschedules_left = 0
                        elif OPTIONS.debug:
                            print(f"no install.log found for recipe R:{recipe.get('id')}")

                        # check for packaging issue (conflicts, missing deps,....)
                        packaging_issues = [
                            {
                                "file": "anaconda.log",
                                "pattern": r".*file conflicts.*installs of ([^ ]*) and ([^ ]*).*",
                                "message": "File conflict between {0} and {1}",
                            },
                            {
                                "file": "packaging.log",
                                "pattern": r".*packaging: ([^ ]*) conflicts with ([^ ]*).*",
                                "message": "File conflict between {0} and {1}",
                            },
                            {
                                "file": "packaging.log",
                                "pattern": r".*WARN packaging: ([^ ]*) requires ([^ ]*)",
                                "message": "Package {0} has missing dependency: {1}",
                            },
                        ]
                        for issue in packaging_issues:
                            logs = LogCache.get_recipe_logs(
                                recipe_id=recipe.get("id"),
                                log_name=issue["file"],
                            )
                            if len(logs) > 0:
                                for line in logs[0].split("\n"):
                                    if match := re.match(issue["pattern"], line):
                                        package_name1 = match.group(1)
                                        package_name2 = match.group(2)
                                        self.status_description = issue["message"].format(
                                            package_name1,
                                            package_name2,
                                        )
                                        # packaging issues are irrecoverable, don't reschedule
                                        self.reschedules_left = 0
                            elif OPTIONS.debug:
                                print(
                                    f"no {issue['file']} found for recipe R:{recipe.get('id')}",
                                )
                    # only report as aborted if last task is aborted
                    # otherwise single task in the middle can make recipe aborted
                    # eventhough it's mostly usable [BZ#1327991]
                    last_task = rs_xml.findall(".//task")[-1]
                    if last_task.get("status") == "Aborted":
                        return False

        # failed /distribution/install
        if "install" not in self.ignore_list:
            for task in rs_xml.findall(".//task"):
                if (task.get("name") == "/distribution/install") and (
                    (task.get("result") not in ("Pass", "Warn", "New"))
                    or (task.get("status") in ("Aborted"))
                ):
                    self.status_description = "installation failed"
                    return False

        # critical task failure
        for task in rs_xml.findall(".//task"):
            if task.get("name") in OPTIONS.critical_tasks and (
                (
                    task.get("result") not in ("Pass", "New")
                    and task.get("status") not in ("Cancelled")
                )
                or task.get("status") in ("Aborted")
            ):
                self.status_description = f"critical task {task.get('name')} failed"
                # irrecoverable error, don't reschedule
                if task.get("name") in {
                    "/distribution/errata/setup",
                    "/distribution/errata/prepare",
                }:
                    self.reschedules_left = 0
                    self.status_description = (
                        f"critical task {task.get('name')} failed (irrecoverable)"
                    )
                return False

        # external watchdog
        if "watchdog" not in self.ignore_list:
            for task in rs_xml.findall(".//result"):
                if task.text == "External Watchdog Expired":
                    if task.getparent().getparent().get("name") == "/distribution/reservesys":
                        if OPTIONS.debug:
                            print(f"ignoring external watchdog in reservesys (RS:{self.id})")

                        continue
                    self.status_description = "killed by external watchdog"
                    return False

        return True

    def get_machines(self):
        """Get list of machines on which this RS ran."""
        return [
            system
            for machine in self.get_xml().findall(".//recipe")
            if (system := machine.get("system")) is not None
        ]

    def get_destructive_tasks(self):
        """Find if there is an aborted task and is followed only by other aborts."""
        rs_xml = self.get_xml()

        first_aborted_task = None
        last_notaborted_task = None
        abort_count = 0
        results = []
        for task in rs_xml.findall(".//task"):
            if task.get("name") not in ("/distribution/install", "/distribution/reservesys"):
                if first_aborted_task is None:
                    if task.get("status") == "Aborted":
                        first_aborted_task = task
                    else:
                        last_notaborted_task = task
                elif task.get("status") != "Aborted":
                    first_aborted_task = None
                    last_notaborted_task = task
                    abort_count = 0
                else:
                    abort_count += 1
        # last_notaborted_task == None => everything is aborted including install => no destrtask
        # first_aborted_task == None => nothing is aborted => no destrtask
        if last_notaborted_task is not None and first_aborted_task is not None:
            # if the task has at least one OK result then it's the culprit. Otherwise
            # it might be its predecessor - the last_notaborted_task
            for result in first_aborted_task.findall(".//result"):
                if not (
                    "LOCALWATCHDOG" in result.get("path")
                    or "Watchdog" in result.text
                    or "Task aborted by task" in result.text
                ):
                    results = [first_aborted_task]
            results = [first_aborted_task, last_notaborted_task]

        # if a destructive task is a setup task then we need to keep it [BZ#1331308]
        new_results = []
        for task in results:
            is_setup = False
            params = task.findall(".//param")
            for param in params:
                if param.get("name") == "WOW_TASK_TYPE" and param.get("value") == "setup":
                    if OPTIONS.debug:
                        print(
                            f"not marking {task.get('name')} as destructive "
                            f"because it's a setup task",
                        )
                    is_setup = True
            if not is_setup:
                new_results.append(task)
        return new_results

    def get_task_info(self):
        """Get the task_info object via XMLRPC."""
        tries = 3
        task_info = None
        signal.signal(signal.SIGALRM, AlarmError.alarm_handler)
        while tries > 0:
            try:
                # Wait up to 60 seconds for xml client to return
                signal.alarm(60)
                task_info = XMLRPC_CLIENT.taskactions.task_info(f"RS:{self.id}")
                # Cancel alarm
                signal.alarm(0)
                break
            except AlarmError:
                if OPTIONS.debug:
                    print("timeout of get_task_info")
            except Exception:
                # Cancel alarm
                signal.alarm(0)
                time.sleep(10)
                tries -= 1
        # restore the default handler
        signal.signal(signal.SIGALRM, signal.SIG_DFL)
        return task_info

    def is_finished(self):
        """Return if the recipeSet is no longer running in Beaker."""
        # unless in aggressive mode, don't fetch the whole xml
        if not OPTIONS.aggressive:
            task_info = self.get_task_info()
            return bool(task_info and task_info["is_finished"])

        # aggressive mode: heavy-weight check using the full xml result
        finished = True
        rs_xml = self.get_xml()
        # check recipeset status
        for recipe in rs_xml.findall(".//recipe"):
            if recipe.get("status") in {
                "Running",
                "Queued",
                "New",
                "Scheduled",
                "Processed",
                "Waiting",
                "Installing",
            }:
                finished = False
        # check all end-tasks in recipeset
        for task in rs_xml.findall(".//task"):
            if (task.get("name") in OPTIONS.end_tasks) and (task.get("result") not in {"New"}):
                finished = True
        if finished and self.is_ok():
            self.status_description = "ok"
        return finished

    def is_nacked(self):
        """Return if the recipeSet is nacked in Beaker."""
        rs_xml = self.get_xml()
        return rs_xml.get("response") == "nak"

    def nack(self):
        """Nack the recipeSet in Beaker."""
        if OPTIONS.debug:
            print(f"nacking RS:{self.id}")
        p = Bkr.run(["job-modify", "-r", "nak", f"RS:{self.id}"])
        if p.wait() != 0:
            print(f"nacking RS:{self.id} failed")
            return False
        return True

    def cancel(self):
        """Cancel the recipeSet in Beaker."""
        if OPTIONS.debug:
            print(f"canceling RS:{self.id}")
        p = Bkr.run(
            [
                "job-cancel",
                f"RS:{self.id}",
                "--msg",
                f"Canceled by beaker-jobwatch: {self.status_description}",
            ],
        )
        if p.wait() != 0:
            print(f"canceling RS:{self.id} failed")
            return False
        return True

    def reschedule(self):
        """Reschedule the recipeSet in Beaker."""
        if self.reschedules_left == 0:
            print("recipeSet reached maximum number of reschedules")
            return -1, None, None
        # download vanilla xml job definition
        try:
            (output, error) = Bkr.run_repeatedly(
                ["job-clone", "--xml", "--dryrun", f"RS:{self.id}"],
            )
        except Bkr.MaxRetriesReachedError:
            print(f"getting job definition of RS:{self.id} for cloning failed")
            return -1, None, None
        parser = lxml.etree.XMLParser(remove_blank_text=True)
        job_xml = lxml.etree.XML(output, parser)

        # update whiteboard
        whiteboard = job_xml.find("whiteboard")
        reason = f"clone of TJ#{self.job}/RS:{self.id}"
        whiteboard.text += f"({reason} #{1 + self.max_reschedules - self.reschedules_left})"

        # set group for the job
        if OPTIONS.job_group:
            job_xml.set("group", OPTIONS.job_group)
            if OPTIONS.debug:
                print(f"setting job group: {OPTIONS.job_group}")

        # set owner for the job
        if OPTIONS.job_owner:
            job_xml.set("user", OPTIONS.job_owner)
            if OPTIONS.debug:
                print(f"setting job owner: {OPTIONS.job_owner}")

        # skip destructive tasks
        if self.skip_destructive:
            for destructive_task in self.get_destructive_tasks():
                for task in job_xml.findall(".//task"):
                    if (task.get("name") == destructive_task.get("name")) and (
                        task.getnext() is not None
                    ):
                        if OPTIONS.debug:
                            print(
                                f"removing destructive task {destructive_task.get('name')} "
                                f"(id: {destructive_task.get('id')})",
                            )
                        task.getparent().remove(task)
        # skip broken machine
        if self.skip_machine:
            for elem_hostrequire in job_xml.findall(".//hostRequires"):
                for machine in self.get_machines():
                    elem_and = lxml.etree.SubElement(elem_hostrequire, "and")
                    elem_hostname = lxml.etree.SubElement(elem_and, "hostname")
                    elem_hostname.set("op", "!=")
                    elem_hostname.set("value", machine)
                    if OPTIONS.debug:
                        print(f"excluding machine: {machine}")

        # dump to a file for submitting
        (fd, filename) = tempfile.mkstemp()
        file = os.fdopen(fd, "wb")
        file.write(lxml.etree.tostring(job_xml))
        file.close()

        # submit to beaker as new job
        try:
            (output, error) = Bkr.run_repeatedly(["job-submit", filename])
            output = output.decode()
            error = error.decode()
        except Bkr.MaxRetriesReachedError:
            print(f"failed to submit cloned RS:{self.id}")
            return -1, None
        if not (match := re.match(r"^Submitted: \['J:([0-9]*)'\]", output)):
            print(f"can't parse job id from job-clone output: {output}", end=" ")
            print(f"job definition: {filename}, stderr: {error}")
            return -1, None, None
        pathlib.Path(filename).unlink()
        job = match.group(1)
        if OPTIONS.debug:
            print(f"new job TJ#{job} scheduled")
        recipe_set = RecipeSetCache.add_job(job)[0]
        if not self.is_finished():
            self.cancel()
        self.nack()
        self.active = False
        return int(job), reason, recipe_set

    def task_statuses(self):
        """How many tasks are in particular status."""
        statuses = {
            "New": 0,
            "Processed": 0,
            "Queued": 0,
            "Scheduled": 0,
            "Waiting": 0,
            "Running": 0,
            "Completed": 0,
            "Cancelled": 0,
            "Aborted": 0,
        }
        for task in self.get_xml().findall(".//task"):
            statuses[task.get("status")] += 1
        return statuses

    def recipe_ids(self):
        """Return list of all recipes that are part of this RS."""
        return [recipe.get("id") for recipe in self.get_xml().findall(".//recipe")]

    def id_string(self) -> str:
        """Return a string representing this RS's ID.

        For a RS with a single recipe, return the recipe ID instead
        because it's more easily converted to URL (via doer).
        """
        if len(self.recipe_ids()) == 1:
            return f"R:{self.recipe_ids()[0]}"
        return f"RS:{self.id}"


def main(args: list[str] | None = None) -> None:
    """Run the beaker-jobwatch script."""
    print(f"beaker-jobwatch v{__version__}")

    usage = """beaker-jobwatch watches beaker jobs and reschedules any recipe-sets that fail because
    of test-environment problems.

    beaker-jobwatch --job 127346[,TJ#1267,3575,...]
    """
    parser = optparse.OptionParser(usage=usage)
    parser.set_defaults(max_reschedules=3)
    parser.set_defaults(sleep=600)
    parser.add_option(
        "-j",
        "--job",
        dest="jobs",
        action="append",
        help="specify job(s) to be investigated",
    )
    parser.add_option(
        "-m",
        "--max",
        dest="max_reschedules",
        metavar="MAX",
        type="int",
        help="Maximum number of times one recipe set will be rescheduled",
    )
    parser.add_option(
        "",
        "--dry-run",
        dest="dry_run",
        action="store_true",
        help="Do not reschedule, cancel or nack failed recipesets, do not update run notes",
    )
    parser.add_option("", "--debug", dest="debug", action="store_true", help="Very verbose logging")
    parser.add_option(
        "",
        "--no-wait",
        dest="no_wait",
        action="store_true",
        help="Don't wait until running jobs finish",
    )
    parser.add_option(
        "",
        "--sleep",
        dest="sleep",
        type="int",
        help="""How long to wait before polling beaker for updated results (default: 600 seconds),\n
    pressing ENTER will skip the pause and force an update""",
    )
    parser.add_option(
        "",
        "--skip-destructive-task",
        dest="skip_destructive",
        action="store_true",
        help="Don't reschedule task(s) suspected of breaking the recipe",
    )
    parser.add_option(
        "",
        "--skip-broken-machines",
        dest="skip_machine",
        action="store_true",
        help="Instruct Beaker to not reuse machine(s) that ran the failed recipe set.",
    )
    parser.add_option(
        "",
        "--load-nacked",
        dest="load_nacked",
        action="store_true",
        help="Don't ignore nacked recipesets in specified jobs",
    )
    parser.add_option(
        "",
        "--ignore-failure",
        dest="ignore_list",
        action="append",
        default=["install"],
        help="Comma separated list of checks that will be ignored, these will not cause jobs to "
        "be canceled and rescheduled. Can be any combination of install, abort, watchdog and "
        "destructive-task or none. Default: install.",
    )
    parser.add_option(
        "",
        "--reboot",
        dest="reboot",
        action="store_true",
        help="Reboot unresponsive machines",
    )
    parser.add_option(
        "",
        "--reboot-time",
        dest="reboot_time",
        type="int",
        default=60,
        help="Reboot time treshold (defaults to 60 minutes)",
    )
    parser.add_option(
        "",
        "--max-reboots-without-ping",
        dest="max_reboots_without_ping",
        type="int",
        default=3,
        help="Reboot count treshold for machines not giving a ping",
    )
    parser.add_option(
        "",
        "--critical-task",
        dest="critical_tasks",
        action="append",
        default=[""],
        help="Failure of this task will trigger recipe cancellation, nacking and reschedule",
    )
    parser.add_option(
        "",
        "--end-task",
        dest="end_tasks",
        action="append",
        default=[""],
        help="Consider recipeset finished if an end task has completed, don't wait for the rest of "
        "the recipeset. Requires aggressive mode.",
    )
    parser.add_option(
        "",
        "--username",
        dest="username",
        help="Username for communication with beaker (instead of kerberos ticket)",
    )
    parser.add_option(
        "",
        "--password",
        dest="password",
        help="Password for communication with beaker (instead of kerberos ticket)",
    )
    parser.add_option(
        "",
        "--job-group",
        dest="job_group",
        help="Reschedule jobs as group jobs belonging to given group",
    )
    parser.add_option(
        "",
        "--job-owner",
        dest="job_owner",
        help="Reschedule jobs as delegated to the given owner",
    )
    parser.add_option(
        "",
        "--aggressive",
        dest="aggressive",
        action="store_true",
        default=False,
        help="Fetch full results repeatedly, not just after recipe finishes. Stresses Beaker a "
        "lot, use with caution!",
    )
    parser.add_option(
        "",
        "--queue-timeout",
        dest="queue_timeout",
        type="int",
        default=None,
        help="Abort recipe set when queued for longer than defined (in seconds). Default: not set",
    )
    global OPTIONS  # pylint: disable=global-statement
    (OPTIONS, _) = parser.parse_args(args)

    if not OPTIONS.jobs:
        parser.print_help()
        sys.exit(1)

    # to evaluate end tasks, we need to fetch tasks results repeatedly
    # so turn aggressive mode on
    if len(OPTIONS.end_tasks) > 1:
        OPTIONS.aggressive = True
    if OPTIONS.debug:
        print(f"options: {OPTIONS}")

    if OPTIONS.debug and OPTIONS.reboot:
        print(f"reboot time: {OPTIONS.reboot_time} min")

    if "BEAKER_USERNAME" in os.environ:
        Bkr.username = os.environ["BEAKER_USERNAME"]
    if "BEAKER_PASSWORD" in os.environ:
        Bkr.password = os.environ["BEAKER_PASSWORD"]
    if OPTIONS.username:
        Bkr.username = OPTIONS.username
    if OPTIONS.password:
        Bkr.password = OPTIONS.password

    ignore_list = []
    if OPTIONS.ignore_list:
        for option in ",".join(OPTIONS.ignore_list).split(","):
            if option not in ("install", "abort", "watchdog", "destructive-task", "none"):
                print(
                    f"Invalid ignore-failure value {option}. Valid values are: "
                    f"install,abort,watchdog,destructive-task",
                )
                sys.exit(1)
        ignore_list = ",".join(OPTIONS.ignore_list).split(",")
        if "none" in ignore_list:
            ignore_list = []

    RecipeSetCache.update_interval = OPTIONS.sleep - 1
    # list of successfully finished recipesets - will constitute results
    herd_finished_ok = []
    # list of unsuccessfully finished recipes
    herd_broken = []

    # dots are printed as a sort of status if nothing changes
    # lets count how many dots have we printed and add line break after 30
    dot_counter = 0

    # print status even if nothing changed since the last iteration, set to
    # True to print status at start
    force_status = True

    # good_recipesets: how many OK recipesets we must have to finish successfully
    # herd: list of recipesets we monitor - not finished yet
    # duration: measure how long does each arch take to finish
    (good_recipesets, herd, duration) = read_options(OPTIONS, ignore_list)

    # main cycle iterating over all recipesets in herd until all are done =
    # = moved to other herds
    while True:
        fingerprint = str([herd, herd_finished_ok, herd_broken])

        # finished successfully
        new_herd = []
        for rs in herd:
            if rs.is_finished() and rs.is_ok():
                if OPTIONS.debug:
                    RecipeSetCache.dump_to_disk(rs.id)
                herd_finished_ok.append(rs)
            else:
                new_herd.append(rs)
        herd = new_herd

        # broken (running or finished, doesn't matter)
        new_herd = []
        for rs in herd:
            if not rs.is_ok():
                if rs.reschedules_left > 0:
                    if not OPTIONS.dry_run:
                        new_job, _reason, rs_id = rs.reschedule()
                        if new_job > 0:
                            if OPTIONS.debug:
                                print(f"watching new recipeSet {rs_id}")
                            herd.append(
                                RecipeSet(
                                    rs_id,
                                    new_job,
                                    rs.reschedules_left - 1,
                                    OPTIONS.max_reschedules,
                                    rs.skip_destructive,
                                    rs.skip_machine,
                                    ignore_list,
                                ),
                            )
                # normally cancellation happens after successful reschedule,
                # here we do it for recipes that won't be rescheduled (e.g. queue timeouts)
                elif not rs.is_finished():
                    rs.cancel()
                    rs.nack()
                    rs.active = False
                if OPTIONS.debug:
                    RecipeSetCache.dump_to_disk(rs.id)
                herd_broken.append(rs)
            else:
                new_herd.append(rs)
        herd = new_herd

        # print status
        status = ""
        new_fingerprint = str([herd, herd_finished_ok, herd_broken])
        if force_status or fingerprint != new_fingerprint:
            force_status = False
            # header
            print("\n")
            status += f"{datetime.datetime.now(tz=datetime.UTC).strftime('%Y-%m-%d %H:%M:%S')}\n"
            whiteboards = []
            for rs in herd + herd_finished_ok + herd_broken:
                whiteboard = RecipeSetCache.whiteboard(rs.id)
                # strip whiteboard of the postfix we add during rescheduling ("clone of ...")
                whiteboard = re.sub(r" *\(clone of.*\) *", "", whiteboard)
                if whiteboard not in whiteboards:
                    status += f"{whiteboard}\n"
                    whiteboards.append(whiteboard)
            # individual herds
            status += f"Broken: {len(herd_broken)}\n"
            for rs in herd_broken:
                # indicate if max_reschedules has been reached
                # hard to spot from log if default (=3) wasn't used
                if rs.reschedules_left == 0:
                    attempts_string = f"{OPTIONS.max_reschedules - rs.reschedules_left} (max)"

                else:
                    attempts_string = str(OPTIONS.max_reschedules - rs.reschedules_left)
                status += (
                    f"\tTJ#{rs.job} {rs.id_string()} ({rs.description}): "
                    f"broken #{attempts_string}: {rs.status_description}\n"
                )

            status += f"Running:   {len(herd)}/{good_recipesets}\n"
            for rs in herd:
                stats = rs.task_statuses()
                tasks_finished = stats["Aborted"] + stats["Cancelled"] + stats["Completed"]
                tasks_all = functools.reduce(lambda x, y: x + y, list(stats.values()))
                # unless in aggressive mode we don't have up2date task info so ignore it
                if OPTIONS.aggressive:
                    tasks_stat_string = f"({tasks_finished}/{tasks_all})"
                else:
                    tasks_stat_string = f"({tasks_all})"
                status += (
                    f"\tTJ#{rs.job} {rs.id_string()} ({rs.description}): "
                    f"running {tasks_stat_string}\n"
                )

            status += f"Completed: {good_recipesets - len(herd)}/{good_recipesets}\n"
            if len(herd) + len(herd_finished_ok) == 0:
                print("no jobs in herd + herd_finished_ok, this should never happen")
                print("(unless you inspect finished broken jobs only)")
                print([f"TJ#{rs.job} {rs.id_string()}" for rs in (herd)])
                print([f"TJ#{rs.job} {rs.id_string()}" for rs in (herd_finished_ok)])
                print([f"TJ#{rs.job} {rs.id_string()}" for rs in (herd_broken)])
            else:
                if len(herd_broken) + len(herd_finished_ok) != 0:
                    status += f"\t{
                        functools.reduce(
                            lambda x, y: f'{x} {y}',
                            {f'TJ#{rs.job}' for rs in (herd_broken + herd_finished_ok)},
                        )
                    }\n"
                status += (
                    f"https://beaker.engineering.redhat.com/matrix/?"
                    f"toggle_nacks_on=on&job_ids={
                        functools.reduce(
                            lambda x, y: f'{x}+{y}',
                            {rs.job for rs in (herd + herd_finished_ok + herd_broken)},
                        )
                    }"
                )
            print(status)

            # check if some arch isn't completely finished, note the time
            for arch, arch_duration in duration.items():
                if arch not in [x.description for x in herd] and arch_duration["end"] is None:
                    arch_duration["end"] = datetime.datetime.now(tz=datetime.UTC)
        else:
            sys.stdout.write(".")
            dot_counter += 1
            if dot_counter == STATUS_DOT_COUNT:
                print(f"\n{datetime.datetime.now(tz=datetime.UTC).strftime('%Y-%m-%d %H:%M:%S')}")
                dot_counter = 0
        sys.stdout.flush()

        if not OPTIONS.no_wait and len(herd) > 0:
            # sleep between updates, hitting ENTER will skip
            # the sleep and force job-results update
            if sys.stdin.isatty():
                slept = 0
                while slept < OPTIONS.sleep:
                    rready, _wready, _xready = select.select([sys.stdin], [], [], 0.0001)
                    for s in rready:
                        if s == sys.stdin:
                            sys.stdin.readline()  # read line = enter pressed
                            print("skipping pause and forcing update of all running recipe-sets")
                            slept = OPTIONS.sleep
                            force_status = True
                            for rs_id in {rs.id for rs in herd}:
                                RecipeSetCache.update(rs_id, force=True)
                            # flush the rest of input so we start clear the next round
                            while len(select.select([sys.stdin.fileno()], [], [], 0.0)[0]) > 0:
                                os.read(sys.stdin.fileno(), 4096)
                                # just in case we've lost tty to not block infinitely
                                if not sys.stdin.isatty():
                                    break
                    time.sleep(1)
                    slept += 1
            else:
                # not running on terminal, just sleep the whole time
                time.sleep(OPTIONS.sleep)
        else:
            break
    # dump herds for parsing out the (un)successful recipe sets
    if OPTIONS.debug:
        print(f"herd_broken: {','.join(str(x) for x in herd_broken)}")
        print(f"herd_finished_ok: {','.join(str(x) for x in herd_finished_ok)}")
    # print duration
    if not OPTIONS.no_wait:
        delta = max(d["end"] for d in duration.values()) - min(
            d["start"] for d in duration.values()
        )
        print(f"duration: {delta}")
        if OPTIONS.debug:
            for arch, arch_duration in duration.items():
                print(f"duration of {arch}: {arch_duration['end'] - arch_duration['start']}")
    if False in [rs.is_ok() for rs in herd if rs.active] or good_recipesets > len(herd_finished_ok):
        print("finished unsuccessfully")
        sys.exit(2)

    print("finished successfully")
